#
# Copyright (c) 1996 Regents of the University of California.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
# 	This product includes software developed by the MASH Research
# 	Group at the University of California Berkeley.
# 4. Neither the name of the University nor of the Research Group may be
#    used to endorse or promote products derived from this software without
#    specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Header: /usr/src/mash/repository/common/Tcl/tcl-object.tcl,v 1.30 1998/01/25 23:57:23 gnguyen Exp $
#

#
# A SplitObject is an OTcl object with a C++ shadow object, i.e., an object
# whose implementation is split across OTcl and C++.  This is the base
# class of all such objects.
#

Class InitObject

# init-all-vars:  initializes all default variables for an object
InitObject instproc init-all-vars {classes} {
	foreach cl $classes {
		if {$cl == "Object"} continue
		# depth first: set vars of ancestors first
		$self init-all-vars "[$cl info superclass]"
		foreach var [$cl info vars] {
			if [catch "$self set $var"] {
				$self set $var [$cl set $var]
			}
		}
	}
}

Class SplitObject -superclass InitObject
SplitObject set id 0

SplitObject instproc init args {
	$self next

	#
	# Emulate Object's args processing
	#  1.  Look for pairs of {-cmd val} in args
	#  2.  If "$self $cmd $val" is not valid then put it in $shadow_args
	#
	set object_args ""
	set shadow_args ""
	for {} {$args != ""} {set args [lrange $args 2 end]} {
		set key [lindex $args 0]
		set val [lindex $args 1]
		if {$val != "" && [string match {-[A-z]*} $key]} {
			set cmd [string range $key 1 end]
			if ![catch "$self $cmd $val"] {
				lappend object_args $key $val
				continue
			}
		}
		eval lappend shadow_args $key $val
	}

	if [catch "$self create-shadow $shadow_args"] {
		error "__FAILED_SHADOW_OBJECT_" ""
	}
}

SplitObject instproc destroy {} {
	$self delete-shadow
	$self next
}

SplitObject proc getid {} {
	$self instvar id
	incr id
	return _o$id
}

SplitObject proc is-class cl {
	if [catch "SplitObject info subclass $cl" v] {
		return 0
	}
	return $v
}

SplitObject instproc unknown args {
	if [catch "$self cmd $args" ret] {
		set cls [$self info class]
		global errorInfo
		set savedInfo $errorInfo
		error "error when calling class $cls: $args" $savedInfo
	}
	return $ret
}

proc new { className args } {
	set o [SplitObject getid]
	if [catch "$className create $o $args" msg] {
		if [string match "__FAILED_SHADOW_OBJECT_" $msg] {
			#
			# The shadow object failed to be allocated.
			# 
			delete $o
			return ""
		}
		global errorInfo
		error "class $className: constructor failed ($msg)" $errorInfo
	}
	return $o
}

proc delete o {
	$o delete_tkvar
	$o destroy
}

#
# Register a C++ compiled-in class with OTcl
# Arrange things so we catch create and destroy
# and manage the underlying C++ object accordingly.
# Classes have structured names with the hierarchy
# delineated by slashes, e.g., a side effect of
# creating class "A/B/C" is to create classes "A"
# and "A/B".	
# This routine invoked by TclClass::bind.
#
SplitObject proc register className {
	set classes [split $className /]
	set parent SplitObject
	set path ""
	set sep ""
	foreach cl $classes {
		set path $path$sep$cl
		if ![$self is-class $path] {
			Class $path -superclass $parent
		}
		set sep /
		set parent $path
	}
}

#
# warn if a class variable not defined
# this is in a separate method so user can nop it
#
# In ns, this error happens for two main reasons:
#
#	1. you bound a variable in C but didn't initialize it in tcl
#		To fix:  put initialization code in tcl/lib/ns-default.tcl
#
#	2. you invoked something which assumed that something else had
#		been built (for example, doing "new Node" without having
#		first done "new Simulator")
#		To fix:  do new Simulator (or whatever).
#
SplitObject instproc warn-instvar item {
	$self instvar issuedWarning
	if ![info exists issuedWarning($item)] {
		set issuedWarning($item) 1
		puts stderr "warning: no class variable $item\n"
		$self instvar SplitObject_issued_undeclared_warning
		if ![info exists SplitObject_issued_undeclared_warning] {
			puts stderr "\tsee tcl-object.tcl in tclcl for info about this warning.\n"
			set SplitObject_issued_undeclared_warning 1
		}
	}
}

#
# Initialize a class instance variable from its
# parent class' member value.  If no such variable,
# call the warning method (which can be overridden).
# We search the class hierarchy for the variable
# until we give up at the SplitObject class.
#
SplitObject instproc init-instvar var {
	set cl [$self info class]
	while { "$cl" != "" && "$cl" != "SplitObject" } {
		foreach c $cl {
			if ![catch "$c set $var" val] {
				$self set $var $val
				return
			}
		}
		set parents ""
		foreach c $cl {
			if { $cl != "SplitObject" && $cl != "Object" } {
				set parents "$parents [$c info superclass]"
			}
		}
		set cl $parents
	}
	$self warn-instvar [$self info class]::$var
}

proc tkerror msg {
	global errorInfo
	puts -nonewline "$msg: "
	puts $errorInfo
	exit 1
}

# somehow tkerror is *evaluated* as 'bgerror', and put 'tkerror' in
# bgerror's body will cause an infinite execution loop... :(
proc bgerror msg {
	global errorInfo
	puts -nonewline "$msg: "
	puts $errorInfo
	exit 1
}

#
# Method hooks to allow using public/private keywords
# instead of instproc to define methods.  For now, we
# don't actually enforce these semantics.  Rather, they
# serve only to document the API (and as a hook for
# the automatic doc generator).
#
Object instproc public args {
	eval $self instproc $args
}

Object instproc private args {
	eval $self instproc $args
}

Object instproc proc.public args {
	eval $self proc $args
}

Object instproc proc.private args {
	eval $self proc $args
}

#tkvar stuff


Object instproc tkvar args {
	foreach var $args {
		if { [llength $var] > 1 } {
			set varname [lindex $var 1]
			set var [lindex $var 0]
		} else {
			set varname $var
		}
		uplevel upvar #0 $self/$var $varname
	}
}

Object instproc tkvarname var {
	return $self/$var
}

Object instproc delete_tkvar { } {
	set fullname [$self tkvarname foo]
	regexp "(.*)foo$" $fullname dummy prefix
	foreach global [info globals "$prefix*"] {
		global $global
		unset $global
	}
}

Object instproc info_tkvar { pattern } {
	set pattern [$self tkvarname $pattern]
	return [info globals $pattern]
}

#
# Backward compat: SplitObjects used to be called TclObjects
#
proc TclObject args {
	return [eval SplitObject $args]
}

#
# Procedure to order a list of SplitObject's.
#
# use as: "lsort -command SplitObjectCompare {list}"
#
proc SplitObjectCompare {a b} {
	set o1 [string range $a 2 end]
	set o2 [string range $b 2 end]
	if {$o1 < $o2} {
		return -1
	} elseif {$o1 == $o2} {
		return 0
	} else {
		return 1
	}
}

#
# Support for a simple, java-like import construct.
#
Class Import

proc import args {
	Import instvar instance_
	if ![info exists instance_] {
		set instance_ [new Import]	    
	}
	eval $instance_ import $args
}

Import instproc find_import_dirs {} {
	global env
	if [info exists env(TCLCL_IMPORT)] {
		set list [split $env(TCLCL_IMPORT) :]
	} else {
		set list .
	}
	set dirs ""
	foreach d $list {
		if [file readable $d/importTable] {
			set dirs "$dirs $d"
		}
	}
	return $dirs
}

Import instproc read_table {} {
	$self instvar table_
	set dirs [$self find_import_dirs]
	if { $dirs == "" } {
		puts stderr "can't find an importTable in current directory"
		puts stderr "or in path from TCLCL_IMPORT environment"
		exit 1
	}
	foreach d $dirs {
		$self read_dir $d
	}
}

Import instproc read_dir dir {
	$self instvar table_
	set f [open $dir/importTable "r"]
	while 1 {
		set l [gets $f]
		if [eof $f] {
			break
		}
		set index [lindex $l 0]
		set fname $dir/[lindex $l 1]
		if {[info exists table_($index)]} {
			if {-1!=[lsearch -exact $table_($index) $fname]} {
				continue
			}
		}
		lappend table_($index) $fname
	}
}

Import instproc import args {
	$self instvar table_ loaded_
	if ![info exists table_] {
		$self read_table
	}
	foreach item $args {
		if ![info exist table_($item)] {
			puts stderr "can't find import entry for $item"
			exit 1
		}
		foreach file $table_($item) {
			if [info exists loaded_($file)] {
				continue
			}
			set loaded_($file) 1
			if ![file readable $file] {
				puts stderr "can't find file $file for $item"
				exit 1
			}
			puts stderr "importing $item from $file"
			uplevel \#0 source $file
		}
	}
}
